import zipfile
import struct
import zlib

ZipFile = zipfile.ZipFile

# This ZipFile decompresses directly to disk.
class STUZipFile(zipfile.ZipFile):

    # Note: Only the following methods are safe to call.
        
    # writestr
    # write
    # namelist
    # read_md5
    # read
    # read_into

    def writestr(self, name, bytes):
        return ZipFile.writestr(self, self.root + name, bytes)

    def writestr_uncompressed(self, name, bytes):
        zi = zipfile.ZipInfo(self.root + name)
        zi.compress_type = zipfile.ZIP_STORED
        ZipFile.writestr(self, zi, bytes)

    def write(self, fn):
        return ZipFile.write(self, fn, self.root + fn)
    
    def getinfo_relative(self, fn):
        try:
            return self.getinfo(self.root + fn)
        except KeyError:
            return None
    
    def namelist(self):

        old_list = ZipFile.namelist(self)
        rv = [ ]

        for fn in old_list:
            if fn.startswith(self.root):
                rv.append(fn[len(self.root):])

        return rv
    
    def __init__(self, name, mode, root):
        ZipFile.__init__(self, name, mode, zipfile.ZIP_DEFLATED)

        # A root directory that is prepended to all of the zip
        # operations. This should include the trailing /.
        self.root = root 

    def read(self, name):
        return ZipFile.read(self, self.root + name)
        
    def read_into(self, name, dest):

        name = self.root + name
        
        if self.mode not in ("r", "a"):
            raise RuntimeError, 'read() requires mode "r" or "a"'
        if not self.fp:
            raise RuntimeError, \
                  "Attempt to read ZIP archive that was already closed"
        zinfo = self.getinfo(name)
        filepos = self.fp.tell()

        self.fp.seek(zinfo.header_offset, 0)

        # Skip the file header:
        fheader = self.fp.read(30)
        if fheader[0:4] != zipfile.stringFileHeader:
            raise zipfile.BadZipfile, "Bad magic number for file header"

        fheader = struct.unpack(zipfile.structFileHeader, fheader)
        fname = self.fp.read(fheader[zipfile._FH_FILENAME_LENGTH])
        if fheader[zipfile._FH_EXTRA_FIELD_LENGTH]:
            self.fp.read(fheader[zipfile._FH_EXTRA_FIELD_LENGTH])

        if fname != zinfo.orig_filename:
            raise zipfile.BadZipfile, \
                      'File name in directory "%s" and header "%s" differ.' % (
                          zinfo.orig_filename, fname)

        df = file(dest, "w")
        
        remaining = zinfo.compress_size
        block_size = 2048

        if zinfo.compress_type == zipfile.ZIP_DEFLATED:
            dc = zlib.decompressobj(-15)
            
        while remaining:
            bytes = self.fp.read(min(remaining, block_size))
            remaining -= len(bytes)
            
            if zinfo.compress_type == zipfile.ZIP_DEFLATED:
                bytes = dc.decompress(bytes)
                if remaining == 0:
                    bytes += dc.decompress('Z') + dc.flush()

            df.write(bytes)

        df.close()

        # import binascii
        # crc = binascii.crc32(file(dest, "r").read())
        # assert crc == zinfo.CRC 
        
        self.fp.seek(filepos, 0)
    
    def read_md5(self, name, md5):

        name = self.root + name

        """Return file bytes (as a string) for name."""
        if self.mode not in ("r", "a"):
            raise RuntimeError, 'read() requires mode "r" or "a"'
        if not self.fp:
            raise RuntimeError, \
                  "Attempt to read ZIP archive that was already closed"
        zinfo = self.getinfo(name)
        filepos = self.fp.tell()

        self.fp.seek(zinfo.header_offset, 0)

        # Skip the file header:
        fheader = self.fp.read(30)
        if fheader[0:4] != zipfile.stringFileHeader:
            raise zipfile.BadZipfile, "Bad magic number for file header"

        fheader = struct.unpack(zipfile.structFileHeader, fheader)
        fname = self.fp.read(fheader[zipfile._FH_FILENAME_LENGTH])
        if fheader[zipfile._FH_EXTRA_FIELD_LENGTH]:
            self.fp.read(fheader[zipfile._FH_EXTRA_FIELD_LENGTH])

        if fname != zinfo.orig_filename:
            raise zipfile.BadZipfile, \
                      'File name in directory "%s" and header "%s" differ.' % (
                          zinfo.orig_filename, fname)
                          

        remaining = zinfo.compress_size
        block_size = 2048

        if zinfo.compress_type == zipfile.ZIP_DEFLATED:
            dc = zlib.decompressobj(-15)
            
        while remaining:
            bytes = self.fp.read(min(remaining, block_size))
            remaining -= len(bytes)
            
            if zinfo.compress_type == zipfile.ZIP_DEFLATED:
                bytes = dc.decompress(bytes)
                if remaining == 0:
                    bytes += dc.decompress('Z') + dc.flush()

            md5.update(bytes)

        self.fp.seek(filepos, 0)

